package com.nokia.domainb.capability.controller;

import com.alibaba.fastjson2.JSON;
import com.google.common.collect.Lists;
import com.nokia.domainb.capability.config.TestConfig;
import com.nokia.domainb.dto.RouteDTO;
import de.siegmar.fastcsv.writer.CsvWriter;
import lombok.extern.slf4j.Slf4j;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.context.annotation.Import;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.util.StopWatch;
import org.springframework.web.client.RestTemplate;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Slf4j
@SpringBootTest
@Import(TestConfig.class)
class DomainControllerTest {
    /**
     * 测试结果保存文件夹
     */
    public static final String TEST_DIR = "test/";
    /**
     * 数据文件路径
     */
    public static final String DATA_PATH = TEST_DIR + "data/10000.txt";
    
    @Autowired
    private RestTemplate template;
    @Autowired
    private ThreadPoolExecutor requestExecutor;
    @Autowired
    private ThreadPoolExecutor offerExecutor;
    @Autowired
    private ThreadPoolTaskScheduler scheduler;

    /**
     * 路由查询并发测试
     */
    @Test
    void testRoute() {
        // 测试环境地址
        String url = "http://192.168.50.3:12091/domainB/api/route/";
        // 正式环境地址
//        String url = "http://192.168.70.130:12091/domainB/api/route/";
        // 结果保存的路径
        String resultPath = TEST_DIR + "result/route/";
        // 创建文件夹
        try {
            Files.createDirectories(Paths.get(resultPath));
        } catch (IOException e) {
            log.error("创建文件夹失败: {}", e.getMessage(), e);
            return;
        }
        resultPath = resultPath + "路由查询" + TestConfig.qps + "个线程10000条数据测试结果.csv";
        request(url);
        writeFile(resultPath);
    }

    /**
     * 欠费接口整合并发测试
     */
    @Test
    void testQryOweFee() {
        // 测试环境地址
        String url = "http://192.168.50.3:12091/domainB/api/acting/qryowefee/";
        // 正式环境地址
//        String url = "http://192.168.70.130:12091/domainB/api/acting/qryowefee/";
        // 结果保存的路径
        String resultPath = TEST_DIR + "qryowefee/";
        resultPath = resultPath + "欠费接口整合" + TestConfig.qps + "个线程10000条数据测试结果.csv";
        // 创建文件夹
        try {
            Files.createDirectories(Paths.get(resultPath));
        } catch (IOException e) {
            log.error("创建文件夹失败: {}", e.getMessage(), e);
            return;
        }
        resultPath = resultPath + "欠费接口整合" + TestConfig.qps + "个线程100条数据测试结果.csv";
        request(url);
        writeFile(resultPath);
    }

    /**
     * 5GSA查询并发测试
     */
    @Test
    void testSa() {
        // 测试环境地址
        String url = "http://192.168.50.3:12091/domainB/api/sa/";
        // 正式环境地址
//        String url = "http://192.168.70.130:12091/domainB/api/sa/";
        // 结果保存的路径
        String resultPath = TEST_DIR + "sa/";
        // 创建文件夹
        try {
            Files.createDirectories(Paths.get(resultPath));
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        resultPath = resultPath + "5GSA查询" + TestConfig.qps + "个线程10000条数据测试结果.csv";
        request(url);
        writeFile(resultPath);
    }

    /**
     * 功能查询并发测试
     */
    @Test
    void testSaAndVolte() {
        // 测试环境地址
        String url = "http://192.168.50.3:12091/domainB/api/product/saAndVolte/";
        // 正式环境地址
//        String url = "http://192.168.70.130:12091/domainB/api/product/saAndVolte/";
        // 合并结果保存的路径
        String resultPath = TEST_DIR + "saAndVolte/";
        // 创建文件夹
        try {
            Files.createDirectories(Paths.get(resultPath));
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        resultPath = resultPath + "功能查询" + TestConfig.qps + "个线程10000条数据测试结果.csv";
        request(url);
        writeFile(resultPath);
    }

    /**
     * 用户资料三户返回并发测试
     */
    @Test
    void testUserState() {
        // 测试环境地址
        String url = "http://192.168.50.3:12091/domainB/api/threePart/userState/";
        // 正式环境地址
//        String url = "http://192.168.70.130:12091/domainB/api/threePart/userState/";
        // 合并结果保存的路径
        String resultPath = TEST_DIR + "userState/";
        // 创建文件夹
        try {
            Files.createDirectories(Paths.get(resultPath));
        } catch (IOException e) {
            e.printStackTrace();
            return;
        }
        resultPath = resultPath + "用户资料三户返回" + TestConfig.qps + "个线程10000条数据测试结果.csv";
        request(url);
        writeFile(resultPath);
    }

    /**
     * 并发请求
     * @param url 接口地址
     */
    private void request(String url) {
        try(Stream<String> lines = Files.lines(Paths.get(DATA_PATH))) {
            // 去重
            List<String> list = lines.distinct().collect(Collectors.toList());
            // 切分数据，每秒 qps 个请求
            List<List<String>> parts = Lists.partition(list, TestConfig.qps);
            // 任务执行时间
            Instant startTime = Instant.now();
            for (List<String> p : parts) {
                startTime = startTime.plusSeconds(1);
                // 添加定时任务
                scheduler.schedule(() -> {
                    // 并发请求
                    CompletableFuture.allOf(p.stream().map(t ->
                            CompletableFuture.runAsync(() -> {
                                RouteDTO dto = new RouteDTO("test", t);
                                Object r;
                                LocalDateTime now = LocalDateTime.now();
                                StopWatch stopWatch = new StopWatch();
                                stopWatch.start();
                                try {
                                    r = template.postForObject(url, dto, Object.class);
                                } catch (Exception e) {
                                    log.error(e.getMessage(), e);
                                    r = e.getMessage();
                                }
                                stopWatch.stop();
                                List<String> data = new ArrayList<>();
                                data.add(now.toString());
                                data.add(String.valueOf((long) Math.ceil(stopWatch.getTotalTimeSeconds())));
                                data.add(String.valueOf(stopWatch.getTotalTimeMillis()));
                                data.add(JSON.toJSONString(dto));
                                data.add(JSON.toJSONString(r));
                                // 结果入队列
                                CompletableFuture.runAsync(() -> TestConfig.queue.offer(data), offerExecutor);
                            }, requestExecutor)).toArray(CompletableFuture[]::new));
                }, startTime);
            }
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
    }

    /**
     * 写文件
     *
     * @param path 写入文件路径
     */
    public void writeFile(String path) {
        ScheduledThreadPoolExecutor scheduledExecutor = scheduler.getScheduledThreadPoolExecutor();
        try (OutputStreamWriter osw = new OutputStreamWriter(Files.newOutputStream(Paths.get(path)),
                "gbk");
             CsvWriter csv = CsvWriter.builder().build(osw)) {
            csv.writeRow("时间", "耗时(s)", "耗时(ms)", "入参", "结果");
            while (scheduler.getActiveCount() > 0
                    || !scheduledExecutor.getQueue().isEmpty()
                    || scheduledExecutor.getCompletedTaskCount() < scheduledExecutor.getTaskCount()
                    || requestExecutor.getActiveCount() > 0
                    || !requestExecutor.getQueue().isEmpty()
                    || requestExecutor.getCompletedTaskCount() < requestExecutor.getTaskCount()
                    || offerExecutor.getActiveCount() > 0
                    || !offerExecutor.getQueue().isEmpty()
                    || offerExecutor.getCompletedTaskCount() < offerExecutor.getTaskCount()
                    || !TestConfig.queue.isEmpty()) {
                List<String> t = TestConfig.queue.take();
                csv.writeRow(t);
                log.info("schedulerCompletedTaskCount: {}, schedulerTaskCount: {}, schedulerQueueSize: {}, schedulerActiveCount: {}, requestCompletedTaskCount: {}, requestTaskCount: {}, requestQueueSize: {}, requestActiveCount: {}, offerCompletedTaskCount: {}, offerTaskCount: {}, offerExecutorQueueSize: {}, OfferActiveCount: {}",
                        scheduledExecutor.getCompletedTaskCount(),
                        scheduledExecutor.getTaskCount(),
                        scheduledExecutor.getQueue().size(),
                        scheduler.getActiveCount(),
                        requestExecutor.getCompletedTaskCount(),
                        requestExecutor.getTaskCount(),
                        requestExecutor.getQueue().size(),
                        requestExecutor.getActiveCount(),
                        offerExecutor.getCompletedTaskCount(),
                        offerExecutor.getTaskCount(),
                        offerExecutor.getQueue().size(),
                        offerExecutor.getActiveCount());
            }
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }
    }
}

